#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#if defined(sun) || defined(__sun)
#  include <sys/filio.h>
#endif
#include <fcntl.h>
#include <termios.h>
#include <math.h>

#include "lx200.h"

int	lx200PortFD;
int	lx200ConnectFlag = 0;

static int readn(int fd, char *ptr, int nbytes);
static int writen(int fd, char *ptr, int nbytes);

test_main()
{
	int	i;
	float	startRA,startDec;
	float	endRA, endDec;

	ConnectLX200("/dev/ttyS0");

	printf("Start...");
	fflush(stdout);

	SetRate(SLEW);

	startRA = GetRA();
	startDec = GetDec();
	printf("Starting RA = %f\tDec = %f\n",startRA,startDec);

	i = SlewToCoords(startRA+1.0,startDec+24.0); 
	while( !CheckCoords(startRA+1.0,startDec+24.0) );
	MySleep(2.0);
	printf("SlewToCoords returned %d\n",i);
	endRA = GetRA();
	endDec = GetDec();
	printf("Midpoint RA = %f\tDec = %f\n",endRA,endDec);

	i = SlewToCoords(startRA,startDec); 
	while( !CheckCoords(startRA,startDec) );
	MySleep(2.0);
	printf("SlewToCoords returned %d\n",i);
	startRA = GetRA();
	startDec = GetDec();
	printf("Ending RA = %f\tDec = %f\n",startRA,startDec);

	printf("end\n");
	fflush(stdout);

	DisconnectLX200();
}

void MySleep(float numSecs)
{
	clock_t				startTime;

	startTime = clock();
	while( (clock() - startTime) <
		(int) (numSecs*((float) CLOCKS_PER_SEC)) );
}

int CheckConnectLX200(void)
{
	char	returnStr[128];
	long	i = -1;

	if(lx200ConnectFlag == 0)
		return 0;

	write(lx200PortFD,"#:GL#",5);
	MySleep(1.0);

	ioctl(lx200PortFD,FIONREAD, &i);
	if(i > 0)
	{
		read(lx200PortFD,returnStr,9);

        {
        char    returnStr[20];
        int numRead;

/*	For setting LX200 to HIGH or LOW precision
        for (;;) {
          write(lx200PortFD,"#:P#",5);
          numRead = readn (lx200PortFD,returnStr,14);
          returnStr[14] = '\0';

          printf ("read=%d\n",numRead);
          printf ("Precision=%s\n",returnStr);

          if (!strncmp(returnStr,"LOW ",4)) break; 
          }
*/
        for (;;) {
	  write(lx200PortFD,"#:GR#",5);
	  numRead = readn(lx200PortFD,returnStr,8);
	  returnStr[7] = '\0';
          /* printf("GetRA: %d read %s\n",numRead,returnStr); */
	  /* If long format we have to read one char more */
          if (returnStr[5] == ':') {
	        read(lx200PortFD,returnStr,8);
		break;
		}
	  /* otherwise force long format */
	  write(lx200PortFD,"#:U#",5);
          }
	}
		return 1;
	}
	else
		return 0;
}
	
void ConnectLX200(char *port)
{
	struct termios		tty;
	char	tempChar;

/*
	printf("Connect port=%s\n",port);
*/
	if(lx200ConnectFlag != 0)
		return;

	/* first, do the connect */
	lx200PortFD = open(port,O_RDWR);
	if(lx200PortFD == -1)
	{
		fprintf(stderr,"Error opening ttyS0 in ConnectLX200!\n");
		return;
	}
	tcgetattr(lx200PortFD,&tty);
	cfsetospeed(&tty, (speed_t) B9600);
	cfsetispeed(&tty, (speed_t) B9600);
	tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;
	tty.c_iflag =  IGNBRK;
	tty.c_lflag = 0;
	tty.c_oflag = 0;
	tty.c_cflag |= CLOCAL | CREAD;
	tty.c_cc[VMIN] = 1;
	tty.c_cc[VTIME] = 5;
	tty.c_iflag &= ~(IXON|IXOFF|IXANY);
	tty.c_cflag &= ~(PARENB | PARODD);
	tcsetattr(lx200PortFD, TCSANOW, &tty);

	/* now, flush the input (read) buffer */
	tcflush(lx200PortFD,TCIOFLUSH);

	/* and finally, send a # to terminate any bullshit */
	write(lx200PortFD,"#",1);
 
	lx200ConnectFlag = 1;
}

void SetRate(int newRate)
{
	if(newRate == SLEW) 
		write(lx200PortFD,"#:RS#",5);
	else if(newRate == FIND) 
		write(lx200PortFD,"#:RM#",5);
	else if(newRate == CENTER) 
		write(lx200PortFD,"#:RC#",5);
	else if(newRate == GUIDE) 
		write(lx200PortFD,"#:RG#",5);
}

void StartSlew(int direction)
{
	if(direction == NORTH)
		write(lx200PortFD,"#:Mn#",5);
	else if(direction == EAST)
		write(lx200PortFD,"#:Me#",5);
	else if(direction == SOUTH)
		write(lx200PortFD,"#:Ms#",5);
	else if(direction == WEST)
		write(lx200PortFD,"#:Mw#",5);
}

void StopSlew(int direction)
{
	if(direction == NORTH)
		write(lx200PortFD,"#:Qn#",5);
	else if(direction == EAST)
		write(lx200PortFD,"#:Qe#",5);
	else if(direction == SOUTH)
		write(lx200PortFD,"#:Qs#",5);
	else if(direction == WEST)
		write(lx200PortFD,"#:Qw#",5);
}

void DisconnectLX200(void)
{
	if(lx200ConnectFlag == 1)
		close(lx200PortFD);
	lx200ConnectFlag = 0;
}

void GetRAlx200(char *returnStr)
{
	int	numRead;

	write(lx200PortFD,"#:GR#",5);

	numRead = readn(lx200PortFD,returnStr,9);
	returnStr[8] = '\0';
/*        printf("GetRA:%d read %s\n",numRead,returnStr); */
}

float GetRA(void)
{
	char	returnStr[10];
	float	raHrs,raMin,raSec;
	float	returnVal;
	int	numRead;

	write(lx200PortFD,"#:GR#",5);

	numRead = read(lx200PortFD,returnStr,9);
	returnStr[9] = '\0';
/*        printf("GetRA:%s\n",returnStr); */

	sscanf(returnStr,"%f:%f:%f#",&raHrs,&raMin,&raSec);

	returnVal = raHrs + raMin/60.0 + raSec/600.0;
	return returnVal;
}

void GetDeclx200(char *returnStr)
{
	float	returnVal;
	int	numRead;

	write(lx200PortFD,"#:GD#",5);

	numRead = readn(lx200PortFD,returnStr,10);
	returnStr[3] = ':';
	returnStr[9] = '\0';
/*        printf("GetDec: %d read, %s\n",numRead,returnStr); */
}

float GetDec(void)
{
	char	returnStr[10];
	float	decDeg,decMin,decSec;
	float	returnVal;
	int	numRead;

	write(lx200PortFD,"#:GD#",5);

	numRead = read(lx200PortFD,returnStr,10);
	returnStr[3] = returnStr[6] = returnStr[9] = '\0';
/*        printf("GetDec:%s\n",returnStr); */

	sscanf(returnStr,"%f",&decDeg);
	sscanf(returnStr+4,"%f#",&decMin);
	sscanf(returnStr+7,"%f#",&decSec);

	if(decDeg >= 0.0)
		returnVal = decDeg + decMin/60.0 + decSec/600.0;
	else
		returnVal = decDeg - decMin/60.0 - decSec/600.0;

	return returnVal;
}

int Slew2Coords(char *newRA, char *newDec)
{
	char	outputStr[32], inputStr[2048];
	int	returnVal;

	sprintf(outputStr,"#:Sr%s",newRA);
	outputStr[12] = '#';
	outputStr[13] = '\0';
/*	printf("Slew2Coords: sendRA=%s\n",outputStr); */
	write(lx200PortFD,outputStr,13);
	read(lx200PortFD,inputStr,1);
	if(inputStr[0] != '1')
	{
		fprintf(stderr,"Error setting object RA in SlewToCoords!\n");
		fprintf(stderr,"Tried to send command string %s\n",outputStr);
		fprintf(stderr,"Exiting...\n");
	/*	exit(0); */
	}

	sprintf(outputStr,"#:Sd%s",newDec);
	if (outputStr[4] == ' ' && outputStr[5] == ' ') {
		outputStr[4] = '+';
		outputStr[5] = '0';
		}
	else if (outputStr[4] == ' ' && outputStr[5] == '-') {
		outputStr[4] = '-';
		outputStr[5] = '0';
		}
	else if (outputStr[4] == ' ' && outputStr[5] != '-') {
		outputStr[4] = '+';
		}
	outputStr[7] = 0xdf;
	outputStr[13] = '#';
	outputStr[14] = '\0';
/*	printf("Slew2Coords: sendDec=%s\n",outputStr); */
	write(lx200PortFD,outputStr,14);
	read(lx200PortFD,inputStr,1);
	if(inputStr[0] != '1')
	{
		fprintf(stderr,"Error setting object Dec in SlewToCoords!\n");
		fprintf(stderr,"Tried to send command string %s\n",outputStr);
		fprintf(stderr,"Exiting...\n");
	/*	exit(0); */
	}

	write(lx200PortFD,"#:MS#",5);
	read(lx200PortFD,inputStr,1);
	if(inputStr[0] == '0')
		returnVal = 0;
	else if(inputStr[0] == '1')
	{
		returnVal = 1;
		inputStr[0] = 0;
		while(inputStr[0] != '#')
			read(lx200PortFD,inputStr,1);
	}
	else
	{
		returnVal = 2;
		inputStr[0] = 0;
		while(inputStr[0] != '#')
			read(lx200PortFD,inputStr,1);
	}

	return returnVal;
}

int SlewToCoords(float newRA, float newDec)
{
	int	raHrs,decDeg,decMin;
	float	raMin;
	char	outputStr[32], inputStr[2048];
	int	returnVal;

	raHrs = (int) newRA;
	raMin = (newRA - (float) raHrs)*60.0;
	sprintf(outputStr,"#:Sr%02d:%04.1f#",raHrs,raMin);
	write(lx200PortFD,outputStr,12);
/*	printf("outputStr=%s\n",outputStr); */
	read(lx200PortFD,inputStr,1);
	if(inputStr[0] != '1')
	{
		fprintf(stderr,"Error setting object RA in SlewToCoords!\n");
		fprintf(stderr,"Tried to send command string %s\n",outputStr);
		fprintf(stderr,"Exiting...\n");
/*		exit(0); */
	}

	if(newDec >= 0.0)
	{
		decDeg = (int) newDec;
		decMin = (int) ((newDec - (float) decDeg)*60.0);
	}
	else
	{
		decDeg = (int) (-newDec);
		decMin = (int) (((-newDec) - (float) decDeg)*60.0);
		decDeg *= -1;
	}
	sprintf(outputStr,"#:Sd%+02d%c%02d#",decDeg,0xdf,decMin);
	write(lx200PortFD,outputStr,11);
/*	printf("outputStr=%s\n",outputStr); */
	read(lx200PortFD,inputStr,1);
	if(inputStr[0] != '1')
	{
		fprintf(stderr,"Error setting object Dec in SlewToCoords!\n");
		fprintf(stderr,"Tried to send command string %s\n",outputStr);
		fprintf(stderr,"Exiting...\n");
/*		exit(0); */
	}

	write(lx200PortFD,"#:MS#",5);
	read(lx200PortFD,inputStr,1);
	if(inputStr[0] == '0')
		returnVal = 0;
	else if(inputStr[0] == '1')
	{
		returnVal = 1;
		inputStr[0] = 0;
		while(inputStr[0] != '#')
			read(lx200PortFD,inputStr,1);
	}
	else
	{
		returnVal = 2;
		inputStr[0] = 0;
		while(inputStr[0] != '#')
			read(lx200PortFD,inputStr,1);
	}

	return returnVal;
}


int	CheckCoords(float desRA, float desDec)
{
	float	raError,decError;

	raError = GetRA() - desRA;
	decError = GetDec() - desDec;

	if( fabs(raError) > 1.0/(15.0*30.0) || fabs(decError) > 1.0/30.0)
		return 0;
	else
		return 1;
}

static int writen(fd, ptr, nbytes)
int fd;
char *ptr;
int nbytes;
{
	int nleft, nwritten;
	nleft = nbytes;
	while (nleft > 0) {
		nwritten = write (fd, ptr, nleft);
		if (nwritten <=0 )
			return(nwritten);
		nleft -= nwritten;
		ptr += nwritten;
		}
	return (nbytes - nleft);
}

static int readn(fd, ptr, nbytes)
int fd;
char *ptr;
int nbytes;
{
	int nleft, nread;
	nleft = nbytes;
	while (nleft > 0) {
		nread  = read (fd, ptr, nleft);
		/* printf("readn: %d read\n", nread); */
		if (nread < 0)
			return(nread);
		else if (nread == 0) break;
		nleft -= nread;
		ptr += nread;
		}
	return (nbytes - nleft);
}
